package anasynt;				// Fichier AnaSynt.java
import java.io.*; import analex.*; import symboles.*; import inter.*;
public class AnaSynt {
    private AnaLex lex;		// analyseur lexical pour cet analyseur
    private UniLex pre_vis;     // pr-vision
    Env sommet = null;		// table de symboles courante
    int pris = 0;		// mmoire utilise pour les dclarations
    public AnaSynt(AnaLex l) throws IOException { lex = l; avance(); }
    void avance() throws IOException { pre_vis = lex.parcourir(); }
    void erreur(String s) { throw new Error("vers la ligne "+lex.ligne+": "+s); }
    void reconnaitre(int t) throws IOException {
	if( pre_vis.cat == t ) avance();
	else erreur("erreur syntaxique");
    }
    public void programme() throws IOException {	// programme -> bloc
	Instr s = bloc();
	int avant = s.nouvetiq(); int apres = s.nouvetiq();
	s.emetetiq(avant); s.prod(avant, apres); s.emetetiq(apres);
    }
    Instr bloc() throws IOException {			// bloc -> { decls instrs }
	reconnaitre('{'); Env precedEnv = sommet; sommet = new Env(sommet);
	decls(); Instr s = instrs();
	reconnaitre('}'); sommet = precedEnv;
	return s;
    }
    void decls() throws IOException {
	while( pre_vis.cat == Cat.BASIQUE ) {		// D -> type ID ;
	    Type t = type(); UniLex unl = pre_vis; reconnaitre(Cat.ID); reconnaitre(';');
	    Id id = new Id((Mot)unl, t, pris);
	    sommet.mettre( unl, id );
	    pris = pris + t.taille;
	}
    }
    Type type() throws IOException {
	Type t = (Type)pre_vis;	// a priori pre_vis.cat == Cat.BASIQUE
	reconnaitre(Cat.BASIQUE);
	if( pre_vis.cat != '[' ) return t;		// T -> basique
	else return dims(t);	// rendre un type tableau
    }
    Type dims(Type t) throws IOException {
	reconnaitre('['); UniLex unl=pre_vis; reconnaitre(Cat.NBRE); reconnaitre(']');
	if( pre_vis.cat == '[' )
	    t = dims(t);
	return new Tableau(((Nbre)unl).valeur, t);
    }
    Instr instrs() throws IOException {
	if ( pre_vis.cat == '}' ) return Instr.Nil;
	else return new Suite(instr(), instrs());
    }
    Instr instr() throws IOException {
	Expr x; Instr s, s1, s2;
	Instr precedInstr;	// temporaire pour Instr.Englobant
	switch( pre_vis.cat ) {
	case ';':
	    avance();
	    return Instr.Nil;
	case Cat.SI:
	    reconnaitre(Cat.SI); reconnaitre('('); x = bool(); reconnaitre(')');
	    s1 = instr();
	    if( pre_vis.cat != Cat.SINON ) return new Si(x, s1);
	    reconnaitre(Cat.SINON);
	    s2 = instr();
	    return new Sinon(x, s1, s2);
	case Cat.TANTQUE:
	    TantQue noeudtq = new TantQue();
	    precedInstr = Instr.Englobant; Instr.Englobant = noeudtq;
	    reconnaitre(Cat.TANTQUE); reconnaitre('('); x = bool(); reconnaitre(')');
	    s1 = instr();
	    noeudtq.init(x, s1);
	    Instr.Englobant = precedInstr; // restaurer Instr.Englobant
	    return noeudtq;
	case Cat.FAIRE:
	    Faire noeudfr = new Faire();
	    precedInstr = Instr.Englobant; Instr.Englobant = noeudfr;
	    reconnaitre(Cat.FAIRE); s1 = instr();
	    reconnaitre(Cat.TANTQUE);
	    reconnaitre('('); x = bool(); reconnaitre(')'); reconnaitre(';');
	    noeudfr.init(s1, x);
	    Instr.Englobant = precedInstr; // restaurer Instr.Englobant
	    return noeudfr;
	case Cat.ROMPRE:
	    reconnaitre(Cat.ROMPRE); reconnaitre(';');
	    return new Rompre();
	case '{':
	    return bloc();
	default:
	    return affecter();
	}
    }
    Instr affecter() throws IOException {
	Instr instr; UniLex t = pre_vis;
	reconnaitre(Cat.ID);
	Id id = sommet.obtenir(t);
	if( id == null ) erreur(t.toString() + " non dclar");
	if( pre_vis.cat == '=' ) {			// I -> id = E ;
	    avance(); instr = new Affect(id, bool());
	}
	else {						// I -> elem = E ;
	    Acces x = deplacement(id);
	    reconnaitre('='); instr = new AffElem(x, bool());
	}
	reconnaitre(';');
	return instr;
    }
    Expr bool() throws IOException {
	Expr x = unis();
	while( pre_vis.cat == Cat.OU ) {
	    UniLex unl = pre_vis; avance(); x = new Ou(unl, x, unis());
	}
	return x;
    }
    Expr unis() throws IOException {
	Expr x = egalite();
	while( pre_vis.cat == Cat.ET ) {
	    UniLex unl = pre_vis; avance(); x = new Et(unl, x, egalite());
	}
	return x;
    }
    Expr egalite() throws IOException {
	Expr x = rel();
	while( pre_vis.cat == Cat.EGAL || pre_vis.cat == Cat.NONEGAL ) {
	    UniLex unl = pre_vis; avance(); x = new Rel(unl, x, rel());
	}
	return x;
    }
    Expr rel() throws IOException {
	Expr x = expr();
	switch( pre_vis.cat ) {
	case '<': case Cat.INFEGAL: case Cat.SUPEGAL: case '>':
	    UniLex unl = pre_vis;    avance(); return new Rel(unl, x, expr());
	default:
	    return x;
	}
    }
    Expr expr() throws IOException {
	Expr x = terme();
	while( pre_vis.cat == '+' || pre_vis.cat == '-' ) {
	    UniLex unl = pre_vis; avance(); x = new Arith(unl, x, terme());
	}
	return x;
    }
    Expr terme() throws IOException {
	Expr x = unaire();
	while( pre_vis.cat == '*' || pre_vis.cat == '/' ) {
	    UniLex unl = pre_vis; avance(); x = new Arith(unl, x, unaire());
	}
	return x;
    }
    Expr unaire() throws IOException {
	if( pre_vis.cat == '-' ) {
	    avance(); return new Unaire(Mot.moins, unaire());
	}
	else if( pre_vis.cat == '!' ) {
	    UniLex unl = pre_vis; avance(); return new Non(unl, unaire());
	}
	else return facteur();
    }
    Expr facteur() throws IOException {
	Expr x = null;
	switch( pre_vis.cat ) {
	case '(':
	    avance(); x = bool(); reconnaitre(')');
	    return x;
	case Cat.NBRE:
	    x = new Constante(pre_vis, Type.Entier);   avance(); return x;
	case Cat.REEL:
	    x = new Constante(pre_vis, Type.Flottant); avance(); return x;
	case Cat.VRAI:
	    x = Constante.Vrai;                        avance(); return x;
	case Cat.FAUX:
	    x = Constante.Faux;                        avance(); return x;
	default:
	    erreur("erreur syntaxique");
	    return x;
	case Cat.ID:
	    String s = pre_vis.toString();
	    Id id = sommet.obtenir(pre_vis);
	    if( id == null ) erreur(pre_vis.toString() + " non dclar");
	    avance();
	    if( pre_vis.cat != '[' ) return id;
	    else return deplacement(id);
	}
    }
    Acces deplacement(Id a) throws IOException {	// X -> [E] | [E] X
	Expr i; Expr w; Expr t1, t2; Expr empl;
	Type type = a.type;
	reconnaitre('['); i = bool(); reconnaitre(']');	// 1er indice, X -> [ E ]
	type = ((Tableau)type).de;
	w = new Constante(type.taille);
	t1 = new Arith(new UniLex('*'), i, w);
	empl = t1;
	while( pre_vis.cat == '[' ) {			// multidimensionnel : X -> [ E ] X
	    reconnaitre('['); i = bool(); reconnaitre(']');
	    type = ((Tableau)type).de;
	    w = new Constante(type.taille);
	    t1 = new Arith(new UniLex('*'), i, w);
	    t2 = new Arith(new UniLex('+'), empl, t1);
	    empl = t2;
	}
	return new Acces(a, empl, type);
    }
}
